//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//      Filename                :               ConfigReader.cpp                        //
//                                                                                      //
//      Author                  :               Raman Dhillon                           //
//                                                                                      //
//      Version                 :               1.0                                     //
//                                                                                      //
//      Location                :               Pervasive Communication Lab             //
//                                              University of Colorado at Boulder       //
//                                                                                      //
//      Change Log                                                                      //
//      --------------------------------------------------------------------------      //
//      Change          Author          Date            Reason                          //
//                                                                                      //
//      --------------------------------------------------------------------------      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////


#include "ConfigReader.h"

ConfigReader* ConfigReader::pConfigInstance = 0; // initialize pointer

ConfigReader* ConfigReader::getConfigInstance(char* dictfile)
{

    if(pConfigInstance == 0)
    {

	pConfigInstance = new ConfigReader(dictfile);

    }


    return pConfigInstance;

}


ConfigReader :: ConfigReader(const char* dictfile)
{
    doc = NULL;
    parser = NULL;
    xmlFile = dictfile; 
    doNamespaces       = false;
    doSchema           = false;
    schemaFullChecking = false;
    errorOccurred = false;

    errorHandler = new DOMShowErrorHandler();
}

int ConfigReader :: initializeXMLReader()
{
    // Initialize the XML4C system
    try
    {
            XMLPlatformUtils::Initialize();

    }

    catch (const XMLException& toCatch)
    {
         XERCES_STD_QUALIFIER cerr << "Error during initialization! :\n"
              << StrX(toCatch.getMessage()) << XERCES_STD_QUALIFIER endl;
         return 1;
    }


    AbstractDOMParser::ValSchemes valScheme = AbstractDOMParser::Val_Auto;

    // Instantiate the DOM parser.
    static const XMLCh gLS[] = { chLatin_L, chLatin_S, chNull };
    DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(gLS);
    parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);

    parser->setFeature(XMLUni::fgDOMNamespaces, doNamespaces);
    parser->setFeature(XMLUni::fgXercesSchema, doSchema);
    parser->setFeature(XMLUni::fgXercesSchemaFullChecking, schemaFullChecking);

    if (valScheme == AbstractDOMParser::Val_Auto)
    {
        parser->setFeature(XMLUni::fgDOMValidateIfSchema, true);
    }
    else if (valScheme == AbstractDOMParser::Val_Never)
    {
        parser->setFeature(XMLUni::fgDOMValidation, false);
    }
    else if (valScheme == AbstractDOMParser::Val_Always)
    {
        parser->setFeature(XMLUni::fgDOMValidation, true);
    }

    // enable datatype normalization - default is off
    parser->setFeature(XMLUni::fgDOMDatatypeNormalization, true);

    // And create our error handler and install it
    parser->setErrorHandler(errorHandler);

    //
    //  Get the starting time and kick off the parse of the indicated
    //  file. Catch any exceptions that might propogate out of it.
    //

    //reset error count first
    errorHandler->resetErrors();

    try
    {
	// reset document pool
        parser->resetDocumentPool();

        doc = parser->parseURI(xmlFile);
    }

    catch (const XMLException& toCatch)
    {
        XERCES_STD_QUALIFIER cerr << "\nError during parsing: '" << xmlFile << "'\n"
              << "Exception message is:  \n"
              << StrX(toCatch.getMessage()) << "\n" << XERCES_STD_QUALIFIER endl;
    
        errorOccurred = true;
    }
    catch (const DOMException& toCatch)
    {
        const unsigned int maxChars = 2047;
        XMLCh errText[maxChars + 1];

        XERCES_STD_QUALIFIER cerr << "\nDOM Error during parsing: '" << xmlFile << "'\n"
           << "DOMException code is:  " << toCatch.code << XERCES_STD_QUALIFIER endl;

        if (DOMImplementation::loadDOMExceptionMsg(toCatch.code, errText, maxChars))
	   XERCES_STD_QUALIFIER cerr << "Message is: " << StrX(errText) << XERCES_STD_QUALIFIER endl;

        errorOccurred = true;
    }
    catch (...)
    {
        XERCES_STD_QUALIFIER cerr << "\nUnexpected exception during parsing: '" << xmlFile << "'\n";
        errorOccurred = true;
    }

    if (errorOccurred)
        return 4;
    else
        return 0;

}


int ConfigReader :: printXMLTree()
{
   //
   //  Extract the DOM tree, get the list of all the elements and report the
   //  length as the count of elements.
   //
	
   if(errorHandler == NULL)
	return -1;

  
   if (errorHandler->getSawErrors())
   {
	XERCES_STD_QUALIFIER cout << "\nErrors occurred, no output available\n" << XERCES_STD_QUALIFIER endl;
        errorOccurred = true;
   }
   else
   {
        if (doc) 
	{
		DOMNode* node = (DOMNode*)doc->getDocumentElement();

	        printNode(node);
	
		for (DOMNode* child = node->getFirstChild(); child != 0; child=child->getNextSibling())
		
	        {
		  // char* name = getNodeName(child);

		   char *name = XMLString::transcode(child->getNodeName());

		   //if(strcmp(name,"DWR")==0)
		   //{
			
		   	//cout << "Name = " << name << endl;
			for (DOMNode* n = child->getFirstChild(); n != 0; n=n->getNextSibling())
			printNode(n);

		   //}


		}

	}

   }

   return 0;
}


char* ConfigReader :: getNodeName(DOMNode* node)
{
        return XMLString::transcode(node->getNodeName());
}


int ConfigReader :: printNode(DOMNode* node)
{

   if(node->getNodeType() == DOMNode::ELEMENT_NODE)
   {
	char *name = XMLString::transcode(node->getNodeName());
        std::cout <<"------------------"<<std::endl;
        std::cout <<"Encountered Element : "<< name <<std::endl;

        std::cout << name << std::endl;
        XMLString::release(&name);

	if(node->hasAttributes()) 
	{
        	// get all the attributes of the node
               	DOMNamedNodeMap *pAttributes = node->getAttributes();
               	int nSize = pAttributes->getLength();
               	std::cout <<"\tAttributes" << std::endl;
               	std::cout <<"\t----------" << std::endl;
                    
		for(int i=0;i<nSize;++i) 
		{
               		DOMAttr *pAttributeNode = (DOMAttr*) pAttributes->item(i);
                       	// get attribute name
                       	char *name = XMLString::transcode(pAttributeNode->getName());

                       	std::cout << "\t" << name << "=";
                       	XMLString::release(&name);

                       	// get attribute type
                       	name = XMLString::transcode(pAttributeNode->getValue());
                       	std::cout << name << std::endl;
                       	XMLString::release(&name);
                }
                	
	}


   }
   
   return 0;

}



ConfigReader :: ~ConfigReader()
{
    //
    //  Delete the parser itself.  Must be done prior to calling Terminate, below.
    //

    if (errorHandler)
    {
	delete errorHandler;
	errorHandler = NULL;
    }
    
    parser->release();
    
    // And call the termination method
    XMLPlatformUtils::Terminate();

}

DOMNode* ConfigReader :: getNode(char* commandName)
{

   DOMNode* pNode = NULL;

   if (doc) 
   {
	DOMNode* node = (DOMNode*)doc->getDocumentElement();

	
	for (DOMNode* child = node->getFirstChild(); child != 0; child=child->getNextSibling())
		
	{

		char *name = XMLString::transcode(child->getNodeName());

		if(strcmp(name,commandName)==0)
		{
			//cout << "Name = " << name << endl;
			pNode = child;
			break;

		}

		XMLString::release(&name);

	}

   }

   return pNode;


}


TrafficInfo* ConfigReader :: buildTrafficScenario(char* scenarioName)
{

	DOMNode* scenarioNode = getNode(scenarioName);
	
	if(scenarioNode == NULL) return NULL; // the traffic scenario does not 
					 // exist in the dictionary

	int noOfUsers = 5;
	int startingSeed = 0;
	unsigned int simDuration = 0;

        double dataRatePerUser = 10 ;              // in kbps
        double hurstParameter = 0.75;

        double meanFileSize = 10.0;                    // in kB
        double stdDevFileSize = 1.0;
	int fileSizeUnits = KBYTES;

        string serverAddr;
/*
        double paretoExp;
        double meanInterArrTime;
        double paretoMinValue;
*/

 
   for (DOMNode* node = scenarioNode->getFirstChild(); node != 0; node=node->getNextSibling())
   {	
	char *name = XMLString::transcode(node->getNodeName());
//        cout <<"------------------"<<endl;
//        cout <<"Encountered Element : "<< name <<endl;

	if(node->hasAttributes()) 
	{
		char* attrname = NULL;
		char* attrvalue = NULL;
       		char* attrunits = NULL;
	 	// get all the attributes of the node
               	DOMNamedNodeMap *pAttributes = node->getAttributes();
               	int nSize = pAttributes->getLength();
            //   	cout <<"\tAttributes" << endl;
            //   	cout <<"\t----------" << endl;
                    
		for(int i=0;i<nSize;++i) 
		{
               		DOMAttr *pAttributeNode = (DOMAttr*) pAttributes->item(i);
                       	// get attribute name
                       	attrname = XMLString::transcode(pAttributeNode->getName());

			if(strcmp(attrname,"value") == 0)
			{
                       		// get attribute type
                       		attrvalue = XMLString::transcode(pAttributeNode->getValue());
			}
			else if (strcmp(attrname, "units") == 0)
			{
				attrunits = XMLString::transcode(pAttributeNode->getValue());

			}
				
                       	XMLString::release(&attrname);



                }
		
		if (strcmp (name, "NO_OF_USERS") == 0) noOfUsers = atoi(attrvalue);
		if (strcmp (name, "H_PARAM") == 0) hurstParameter = atof(attrvalue);               	
		if (strcmp (name, "DATA_RATE_PER_USER") == 0) dataRatePerUser = atoi(attrvalue);               	
		
		if (strcmp (name, "MEAN_FILE_SIZE") == 0) 
		{
			meanFileSize = atof(attrvalue);  

			if(strcmp(attrunits,"kB") == 0)	fileSizeUnits = KBYTES;
			if(strcmp(attrunits,"MB") == 0) fileSizeUnits = MBYTES;
			if(strcmp(attrunits, "GB") == 0) fileSizeUnits = GBYTES;
		}
		
		if (strcmp (name, "STD_DEV_FILE_SIZE") == 0) stdDevFileSize = atof(attrvalue);               	
		if (strcmp (name, "SIM_DURATION") == 0) simDuration = atoi(attrvalue);  
		if (strcmp (name, "STARTING_SEED") == 0) startingSeed = atoi(attrvalue);  
		if (strcmp (name, "SERVER") == 0) serverAddr.assign(attrvalue);               	
		
/*		
		if (strcmp (name, "MEAN_INTER_ARR_TIME") == 0) meanInterArrTime = atof(attrvalue);               	
		if (strcmp (name, "MIN_INTER_ARR_TIME") == 0) paretoMinValue = atof(attrvalue);               	
		if (strcmp (name, "PARETO_EXP") == 0) paretoExp = atof(attrvalue);               	
*/
	
		XMLString::release(&attrvalue);
		XMLString::release(&attrunits);

        	XMLString::release(&attrname);
        }
	XMLString::release(&name);
   }

   TrafficInfo* trafInfo = new TrafficInfo(noOfUsers,hurstParameter,dataRatePerUser,meanFileSize, stdDevFileSize, fileSizeUnits, serverAddr, simDuration, startingSeed);


   return trafInfo; 	
}

DOMShowErrorHandler::DOMShowErrorHandler() :

    fSawErrors(false)
{
}

DOMShowErrorHandler::~DOMShowErrorHandler()
{
}


// ---------------------------------------------------------------------------
//  DOMShowHandlers: Overrides of the DOM ErrorHandler interface
// ---------------------------------------------------------------------------
bool DOMShowErrorHandler::handleError(const DOMError& domError)
{
    fSawErrors = true;
    if (domError.getSeverity() == DOMError::DOM_SEVERITY_WARNING)
        XERCES_STD_QUALIFIER cerr << "\nWarning at file ";
    else if (domError.getSeverity() == DOMError::DOM_SEVERITY_ERROR)
        XERCES_STD_QUALIFIER cerr << "\nError at file ";
    else
        XERCES_STD_QUALIFIER cerr << "\nFatal Error at file ";

    XERCES_STD_QUALIFIER cerr << StrX(domError.getLocation()->getURI())
         << ", line " << domError.getLocation()->getLineNumber()
         << ", char " << domError.getLocation()->getColumnNumber()
         << "\n  Message: " << StrX(domError.getMessage()) << XERCES_STD_QUALIFIER endl;

    return true;
}

void DOMShowErrorHandler::resetErrors()
{
    fSawErrors = false;
}
