//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//      Filename                :               Simulator.h	                        //
//                                                                                      //
//      Author                  :               Raman Dhillon                           //
//                                                                                      //
//      Version                 :               1.0                                     //
//                                                                                      //
//      Location                :               Pervasive Communication Lab             //
//                                              University of Colorado at Boulder       //
//                                                                                      //
//      Change Log                                                                      //
//      --------------------------------------------------------------------------      //
//      Change          Author          Date            Reason                          //
//                                                                                      //
//      --------------------------------------------------------------------------      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////


#ifndef _SIMULATOR_H
#define _SIMULATOR_H

#include <iostream>
#include <vector>
#include <string>

#include "User.h"
#include "Barrier.h"
#include "TrafficInfo.h"

using namespace std;

class Simulator
{

 int noOfUsers;
 int dataRatePerUser;
 double hurstParam;
 string serverAddr;

 double meanFileSize;   // in kB
 double stdDevFileSize; // in kB


 vector <User*> userVector;
 Barrier* barr;
 TrafficInfo* trafInfo;
 UserStats* finalUserStats;


 public:

	Simulator(TrafficInfo* traf) : trafInfo(traf)
	{

		if(trafInfo)
		{
			noOfUsers = trafInfo->getNoOfUsers();
		}

   		barr = new Barrier(noOfUsers); 
   
   		trafInfo->showTrafficChars();

		finalUserStats = new UserStats(noOfUsers);

		cout << "\n--------------------------------------------------------------------------------------------------------------------";
   		for (int i=0; i<noOfUsers; i++)
   		{
   			cout << "\nCreating user " << i;

 			userVector.push_back(new User(i, barr, trafInfo, finalUserStats));

   		}
		cout << "\n--------------------------------------------------------------------------------------------------------------------";
	}



	~Simulator() 
	{

 		for (int i=0; i<userVector.size(); i++)
   		{
   			delete userVector[i];
			userVector[i] = NULL;
   		}
 
		int fileSizeUnits = trafInfo->getFileSizeUnits();

		string units;

        	double bytesRecv = finalUserStats->getBytesRecv();

        	switch(fileSizeUnits)
        	{
                	case KBYTES:
                                units = "kB";
                                bytesRecv = bytesRecv / 1000;
                                break;

                	case MBYTES:
                                units = "MB";
                                bytesRecv = bytesRecv / 1000000;
                                break;

                	case GBYTES:
                                units = "GB";
                                bytesRecv = bytesRecv / 1000000000;
                                break;
        	}


		if(finalUserStats)
		{
			
		cout << "\n--------------------------------------------------------------------------------------------------------------------";
		cout << "\nUsers = " << finalUserStats->getUserId() << " \t" ;
        	cout << "Duration = " << finalUserStats->getTestDuration() << " sec\t";
        	cout << "Total download = " << bytesRecv << " " << units << "\t\t";
        	cout << "Aggregate rate = " << finalUserStats->getDownloadRate() << " kbps";

		cout << "\n--------------------------------------------------------------------------------------------------------------------\n\n";
      

		delete finalUserStats;
		finalUserStats = NULL;

		}
   
		if(barr)
   		{
			delete barr;
			barr = NULL;
   		}

   		if(trafInfo)
   		{
			delete trafInfo;
   		}
	}

	void simulateUsers()
	{

		finalUserStats->recordStartTime();

   		for (int i=0; i<userVector.size(); i++)
   		{
   			userVector[i]->simulateUser();
   		}
	}

	void stopSimulation()
	{
   		for (int i=0; i<userVector.size(); i++)
   		{
   			userVector[i]->terminateSimulation();
			cout << "\nStopped user " << i;
   		}

		finalUserStats->recordEndTime();
	}
};

#endif
