//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//      Filename                :               UserThread.cpp                          //
//                                                                                      //
//      Author                  :               Raman Dhillon                           //
//                                                                                      //
//      Version                 :               1.0                                     //
//                                                                                      //
//      Location                :               Pervasive Communication Lab             //
//                                              University of Colorado at Boulder       //
//                                                                                      //
//      Change Log                                                                      //
//      --------------------------------------------------------------------------      //
//      Change          Author          Date            Reason                          //
//                                                                                      //
//      --------------------------------------------------------------------------      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////


#include "UserThread.h"

bool UserThread :: simulate = true;

pthread_mutex_t UserThread :: sig_mutex = PTHREAD_MUTEX_INITIALIZER;

UserThread ::UserThread(int userid, Barrier* b, TrafficInfo* trafInfo, UserStats* userstats)
   : userID(userid), barr(b), trafficInfo(trafInfo), userStats(userstats)
{ 

	URL = trafficInfo->getServerURL();

	http = new HTTPHandler();

	double paretoExp = trafficInfo->getParetoExp();
	double paretoMinValue = trafficInfo->getParetoMinValue();

	int startingSeed = trafficInfo->getStartingSeed();

	startingSeed = startingSeed + userID + startingSeed * userID;	// some operation to produce different seeds
									// not random seeds though, to be able to repeat
									// traffic patterns 

	intArrvDist = new ParetoDist(paretoExp,paretoMinValue,MIN, startingSeed); // userID acts as the seed

	double mean = trafficInfo->getMeanFileSize();
	double stdDev = trafficInfo->getStdDevFileSize();

	fileSizeDist = new LogNormalDist(mean, stdDev, userID);		// userID acts as the seed

}

UserThread :: ~UserThread()
{
	if(http)
	{
		delete http;
		http = NULL;
	}

	if(intArrvDist)
	{
		delete intArrvDist;
		intArrvDist = NULL;
	}

	if(fileSizeDist)
	{
		delete fileSizeDist;
		fileSizeDist = NULL;
	}
}


void UserThread :: execute()
{
       	cout << "\nEntering User " << userID << " Thread Execution...";

	cout << "\nUser " << userID << " waiting at Barrier now...";

	double interArrivalPeriod = 0.0;
	double fileSize = 0.0;
	double reqProcessingTime = 0.0;
	struct timeval reqStartTime;
	struct timeval reqStopTime;

	char fileName[10];
	string units; 

	int fileSizeUnits = trafficInfo->getFileSizeUnits();
	switch(fileSizeUnits) 
	{
		case KBYTES:
				units = "kB";
				break;

		case MBYTES:
				units = "MB";
				break;

		case GBYTES:
				units = "GB";
				break;
	} 


	time_t rawTime;
  	struct tm * timeInfo;
  	char timeStamp[80];

	barr->wait();

	cout << "\nUser " << userID << " Barrier released...";

	userStats->recordStartTime();	// Start time of the user

	// First request of the user, users start activity at different times 
	interArrivalPeriod = intArrvDist->generateRandomVar() * 1000000;
	cout << "\nUser " << userID << " first request starts in " << interArrivalPeriod/1000000 << " sec" ;
	usleep((unsigned long)interArrivalPeriod); 
	
	userStats->setInterArrTime(interArrivalPeriod / 1000000);

	for(;;)
	{
	
		if(checkSimulationStatus() == 0) 
		{
			cout << "\nExiting User";
			break;
	
		}

		cout << "\nUser " << userID;

		interArrivalPeriod = intArrvDist->generateRandomVar() * 1000000;
		fileSize = fileSizeDist->generateRandomVar();


		// convert the file size in kB into a string
		// concatenate this string into the URL requested from the server		
		if(fileSizeUnits != GBYTES)
		{
			sprintf(fileName, "%.f",fileSize);
		}
		else
		{
			if(fileSize <= 0.25) strcpy(fileName,"0_25");
			else if(fileSize <= 0.50) strcpy(fileName,"0_50");
			else if(fileSize <= 0.75) strcpy(fileName,"0_75");
			else if(fileSize <= 1.00) strcpy(fileName,"1_00");
			else if(fileSize <= 1.25) strcpy(fileName,"1_25");
			else if(fileSize <= 1.50) strcpy(fileName,"1_50");
			else if(fileSize <= 1.75) strcpy(fileName,"1_75");
			else strcpy(fileName,"2_00");
	
		//	sprintf(fileName, "%.2f",fileSize);
		}

		string url = URL + fileName;

		url += units;
		url += ".html";
  		
		time (&rawTime);
  		timeInfo = localtime(&rawTime);

  		strftime (timeStamp,80,"%I:%M:%S %p",timeInfo);

  		cout << ", Time=" << timeStamp;
		cout << ", Reqd file=" << fileSize << " " << units;
		cout << ", Inter-arr=" << interArrivalPeriod / 1000000  << " sec";	
		cout << ", URL=" << url;
 
		// check time;	
		gettimeofday(&reqStartTime, NULL);

		userStats->setInterArrTime(interArrivalPeriod/1000000);
		userStats->setFileSize(fileSize);
		userStats->incrNoOfReqSent();
  
		int noOfBytes = http->sendHTTPGetFile((char*)url.c_str());
  	//	http->sendHTTPGet((char*)URL.c_str());
	
		userStats->incrNoOfRespRecv();
		userStats->updateBytesRecv(noOfBytes);
	

	
		gettimeofday(&reqStopTime, NULL);
		// again check time;

		if (fileSizeUnits == KBYTES ||fileSizeUnits == MBYTES )

		{
        	
			if(fileSizeUnits ==KBYTES)
				cout << ", File size=" << noOfBytes / 1000 << " kB";
			else
				cout << ", File size=" << noOfBytes / 1000000 << " MB";

			reqProcessingTime = (reqStopTime.tv_sec - reqStartTime.tv_sec) * 1000 + (reqStopTime.tv_usec - reqStartTime.tv_usec) / 1000;
			cout << ", ReqPr Time=" << reqProcessingTime << " msec";

		} 

		else {
			cout << ", File size=" << noOfBytes / 1000000000 << " GB";

			reqProcessingTime = (reqStopTime.tv_sec - reqStartTime.tv_sec) + (reqStopTime.tv_usec - reqStartTime.tv_usec) / 1000000;
			cout << ", ReqPr Time=" << reqProcessingTime << " sec";
		}

		if(reqProcessingTime < interArrivalPeriod)
		{
			usleep((unsigned long)(interArrivalPeriod-reqProcessingTime));
		}


	}
	
	userStats->recordEndTime();

	barr->wait();
       	cout << "Leaving User " << userID << " Thread Execution...\n";
};


void UserThread :: collectFinalStats()
{
	userStats->recordEndTime();
}


void UserThread :: terminateSimulation()	
{
	pthread_mutex_lock( &sig_mutex );
	simulate = false;
        pthread_mutex_unlock(&sig_mutex);
}

int UserThread :: checkSimulationStatus()
{
	int res;
	
	pthread_mutex_lock( &sig_mutex );

	if(simulate) res = 1;
	else res = 0;

        pthread_mutex_unlock(&sig_mutex);

	return res;

}

